import argparse
from typing import *
from pathlib import Path

from fastapi import FastAPI, HTTPException
from fastapi.responses import FileResponse, RedirectResponse
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
import uvicorn
import junevis.path_fixes as pf

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("--port", default=8000, type=int, help="Port to run the app. ")
parser.add_argument("--dist", default=pf.DIST, type=str, help="Path to the dist folder containing HTML+JS+CSS")

args, _ = parser.parse_known_args()

app = FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# ======================================================================
## Simple Static File Server ##
# ======================================================================
@app.get("/")
async def index():
    """For local development, serve the index.html in the dist folder"""
    return RedirectResponse(url="/index.html")

# the `file_path:path` says to accept any path as a string here. Otherwise, `file_paths` containing `/` will not be served properly
@app.get("/{file_path:path}")
async def send_static_client(file_path:str):
    """ Serves (makes accessible) all files from ./client/ to ``/client/{path}``. Used primarily for development. NGINX handles production.

    Args:
        path: Name of file in the client directory
    """
    f = str(args.dist / file_path)
    print("Finding file: ", f)
    return FileResponse(f)

@app.exception_handler(HTTPException)
async def notfound_exception_handler(request, exc):
    """Catch all response, redirect to index. Necessary when using Vue Router"""
    return RedirectResponse(url="/index.html")

def run():
    uvicorn.run("junevis.server:app", host='127.0.0.1', port=args.port)

if __name__ == "__main__":
    run()
    



"""Fix the paths to important folders in this repository."""

from pathlib import Path
import os

JUNEVIS = Path(
    os.path.abspath(__file__)
).parent
ROOT = JUNEVIS.parent  # Root directory of the github project
CLIENT = JUNEVIS / "client"
PUBLIC = CLIENT / "public"
DIST = CLIENT / "dist"
DEMO = PUBLIC / "demo"
PROJECTS = DEMO / "projects"
DEFAULT_GEOJSON = DEMO / "coxs_bazar.geojson"
AVAILABLE_PROJECTS = DEMO / "availableProjects.txt"


from pathlib import Path
from typing import *
import numpy as np
import pandas as pd
import tables
import logging

logger = logging.getLogger(__name__)


class RecordReader:
    def __init__(self, record_file: Union[Path, str], summary_csv: Optional[Union[Path, str]] = None):
        self.record_file = Path(record_file)

        if summary_csv is not None:
            self.summary_csv = Path(summary_csv)
            self.regional_summary = self.get_regional_summary(
                self.summary_csv
            )
            self.world_summary = self.regional_summary.drop(columns="region") \
                .groupby("time_stamp") \
                .agg(self.aggregator)

    @classmethod
    def from_results_path(cls, results_path:Union[Path, str]="results", record_file: str="june_record.h5", summary_name: Optional[str]=None ):
        results_path = Path(results_path)
        record_file = results_path / record_file

        if summary_name == None:
            summary_csv = None
        else:
            if len(list(results_path.glob(summary_name))) > 0:
                summary_csv = results_path / summary_name
            else:
                raise ValueError(f"Could not find {summary_name} inside of {results_path}")

        return cls(record_file, summary_csv)

    def decode_bytes_columns(self, df):
        str_df = df.select_dtypes([np.object])
        for col in str_df:
            df[col] = str_df[col].str.decode("utf-8")
        return df

    def get_regional_summary(self, summary_path):
        df = pd.read_csv(summary_path)
        self.aggregator = {
            col: np.mean if "current" in col else sum for col in df.columns[2:]
        }
        df = df.groupby(["time_stamp", "region"], as_index=False).agg(self.aggregator)
        df.set_index("time_stamp", inplace=True)
        df.index = pd.to_datetime(df.index)
        return df

    def table_to_df(
        self, table_name: str, index: str = "id", fields: Optional[Tuple] = None
    ) -> pd.DataFrame:
        # TODO: include fields to read only certain columns
        with tables.open_file(self.record_file, mode="r") as f:
            table = getattr(f.root, table_name)
            df = pd.DataFrame.from_records(table.read(), index=index)
        df = self.decode_bytes_columns(df)
        return df

    def get_geography_df(self,):
        areas_df = self.table_to_df("areas")
        super_areas_df = self.table_to_df("super_areas")
        regions_df = self.table_to_df("regions")

        geography_df = areas_df[["super_area_id", "name"]].merge(
            super_areas_df[["region_id", "name"]],
            how="inner",
            left_on="super_area_id",
            right_index=True,
            suffixes=("_area", "_super_area"),
        )
        geography_df = geography_df.merge(
            regions_df, how="inner", left_on="region_id", right_index=True,
        )
        return geography_df.rename(
            columns={geography_df.index.name: "area_id", "name": "name_region"}
        )

    def get_table_with_extras(
        self, table_name, index, with_people=True, with_geography=True
    ):
        logger.info(f"Loading {table_name} table")
        df = self.table_to_df(table_name, index=index)
        if with_people:
            logger.info(f"Loading population table")
            people_df = self.table_to_df("population", index="id")
            logger.info(f"Merging infection and population tables")
            df = df.merge(people_df, how="inner", left_index=True, right_index=True)
            if with_geography:
                logger.info(f"Loading geography table")
                geography_df = self.get_geography_df()
                logger.info(f"Mergeing infection and geography tables")
                df = df.merge(
                    geography_df.drop_duplicates(),
                    left_on="area_id",
                    right_index=True,
                    how="inner",
                )
        if "timestamp" in df.columns:
            df["timestamp"] = pd.to_datetime(df["timestamp"])
        return df


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_Create Project.ipynb (unless otherwise specified).

__all__ = ['init_available_projects', 'summarize_h5', 'pgrid_to_run_parameters', 'collect_statistics', 'fix_geojson',
           'main']

# Cell
from pathlib import Path
import shutil
import os
import pandas as pd
import geopandas as gpd
import shapely
from shapely.geometry.polygon import Polygon
from shapely.ops import cascaded_union, unary_union
import numpy as np
from time import time
from typing import *
import junevis.path_fixes as pf
import json

import junevis.process_loggers as process_loggers

# Cell
def init_available_projects(project_name: str):
    pf.AVAILABLE_PROJECTS.touch()

    with open(str(pf.AVAILABLE_PROJECTS), 'r+') as fp:
        available_projects = set([p.strip() for p in fp.readlines()])
        if project_name in available_projects:
            if not force_add_project:
                raise ValueError(f"Cannot create project of name '{project_name}': Project already exists in {pf.AVAILABLE_PROJECTS}"
    )
            else:
                shutil.rmtree(outdir) # Delete existing project of that name
                fp.truncate(0); fp.seek(0); # Delete file contents
                available_projects.remove(project_name)
        return available_projects

# Cell
def summarize_h5(record_f, outdir):
    """Dependent on the context variable `output_dir`. The actual summarized output is much smaller than the record file itself"""
    start = time()
    runId = record_f.stem.split("_")[1]
    print(f"Processing {runId}: ")
    df = process_loggers.regional_outputs(record_f)

    # Add cumulative columns
    region_grouped_df = df.groupby(level=0)
    df['currently_dead'] = region_grouped_df.deaths.cumsum()
    df['currently_recovered'] = region_grouped_df.recovered.cumsum()

    # Rename region
    df = df.rename_axis(index=["region", "timestamp"])

    outfile = outdir / f"summary_{int(runId):03}.csv"
    print(f"Saving to {str(outfile)}")
    df.to_csv(str(outfile))
    print(f"\nTook {time() - start} seconds")
    print("\n-------\n")
    return df

# Cell
def pgrid_to_run_parameters(parameter_grid: dict) -> dict:
    """Convert parameter_grid dictionary to desired metadata dictionary"""
    run_parameters = {}

    # Create run_parameters
    for k, v in parameter_grid.items():
        for i in range(len(v)):
            curr = run_parameters.get(str(i), {})
            curr[k] = v[i]
            run_parameters[str(i)] = curr

    params_varied = list(parameter_grid.keys())

    return {
        "parameters_varied": params_varied,
        "run_parameters": run_parameters,
    }

# Cell
def collect_statistics(project: Union[str, Path]):
    project = Path(project)
    csvfs = list(project.glob("summary*.csv"))
    dfs = [pd.read_csv(csvf) for csvf in csvfs]
    big_df = pd.concat(dfs, ignore_index=True)

    all_regions = list(set(big_df.region))
    all_fields = [f for f in big_df.columns if f != "Unnamed: 0"]
    all_timestamps = list(set(big_df.timestamp))

    string_fields = ["timestamp", "region", "Unnamed: 0"]
    numerical_fields = [f for f in all_fields if f not in string_fields]
    big_df_num = big_df.loc[:, numerical_fields]
    max_vals = big_df_num.max(axis=0)
    min_vals = big_df_num.min(axis=0)

    df_minmax = pd.DataFrame([max_vals, min_vals], index=["max", "min"])
    field_minmaxes = df_minmax.to_dict(orient="dict")

    return {
        "all_regions": sorted(all_regions),
        "all_timestamps": sorted(all_timestamps),
        "all_fields": sorted(all_fields),
        "field_statistics": field_minmaxes
    }

# Cell
def fix_geojson(gjson_file):
    gdf = gpd.read_file(gjson_file)

    for i, shape in enumerate(gdf.geometry):
        # To reduce the shape of the multipolygon, take the shape of the largest area
        if shape.geom_type == "MultiPolygon":
            polygon = shape[np.argmax(np.array([p.area for p in shape]))]
        else:
            polygon = shape
        gdf.geometry[i] = polygon

    # The frontend operates with a `SSID` field instead of a `region` field to name each area.
    gdf['SSID'] = gdf['region']
    return gdf

# Cell
from fastcore.script import *

@call_parse
def main(record_path:Param("Path to JUNE simulation records and parameter grid", str),
         force_add_project:Param("Overwrite project if it already exists", store_true)=False,
         test_only:Param("Test behavior without changing files", store_true)=False,
         project_name:Param("Name the project. If not provided, use folder name of `record_path`", str)=None,
         description:Param("Description of project", str)="NA",
        ):
    """Create a project that can be visualized from the record files"""

    base = Path(record_path) # Path where loggers and parameter grid are stored
    project_name = base.stem if project_name is None else project_name
    output_dir = pf.PROJECTS / project_name
    if not output_dir.exists() and not test_only: output_dir.mkdir(parents=True)

    active_projects = init_available_projects(project_name)

    record_names = sorted(list(base.glob("*.h5")))
    for r in record_names:
        print(f"Summarizing {r}")
        if not test_only: df = summarize_h5(r, output_dir)

    print("ALL SUMMARIES COMPLETED\n-------------\n-------------\n")

    # Once the summary files have been created, we can accumulate the statistics into the `metadata.json` file
    print("Creating metadata...")
    with open(base / "parameter_grid.json") as fp:
        parameter_grid = json.load(fp)
    param_info = pgrid_to_run_parameters(parameter_grid)
    project_stats = collect_statistics(output_dir)

    # Now we can save the metadata for this project, including the optional description
    metadata = {"description": description}; [metadata.update(p) for p in [param_info, project_stats]];
    if not test_only:
        with open(output_dir / "metadata.json", 'w+') as fp:
            json.dump(metadata, fp, indent=4)

    # Copy over the geography description
    print("Fixing geojson...")
    gdf = fix_geojson(base / "sites.geojson")
    if not test_only: gdf.to_file(output_dir / "sites.new.geojson", driver='GeoJSON')

    # Add to available projects
    print(f"Adding '{project_name}' to {pf.AVAILABLE_PROJECTS}")
    new_available_projects = "".join(["\n" + p for p in (list(active_projects) + [project_name])]).strip()
    print(f"New projects: {new_available_projects}")

    if not test_only:
        with open(pf.AVAILABLE_PROJECTS, 'r+') as fp:
            fp.write(new_available_projects)

    print("COMPLETE")


import pandas as pd
from pathlib import Path
import datetime
import time
from junevis.record_reader import RecordReader
from typing import *

def split_by_age(df, age_bins, out_column_name, group_on=["name_region", "timestamp"]):
    df_by_age = df.groupby(
        [pd.cut(df["age"], bins=age_bins, right=False)] + group_on,
    ).size()
    df_by_age = df_by_age.unstack(level=0, fill_value=0)
    df_by_age.columns.name = None
    df_by_age.columns = [
        out_column_name
        + "_"
        + "_".join(i.replace("[", "").replace(")", "").split(", "))
        for i in df_by_age.columns.astype(str)
    ]
    return df_by_age

def get_infection_locations(infections_df):
    n_infections_by_location = (
        infections_df.groupby(["name_region", "timestamp", "location_specs"])
        .size()
        .unstack(fill_value=0)
    )
    n_infections_by_location.columns = [
        "n_infections_in_" + col for col in n_infections_by_location.columns
    ]
    n_infections_by_location.columns.name = None
    return n_infections_by_location

def get_regional_outputs(df, age_bins, column_name):
    daily = df.groupby(["name_region", "timestamp"]).size()
    by_age = split_by_age(df, age_bins, column_name)
    pd.testing.assert_series_equal(
        daily.groupby("name_region").sum(),
        by_age.sum(axis=1).groupby("name_region").sum(),
        check_dtype=False,
    )
    return pd.DataFrame(daily.rename(column_name)).merge(
        by_age, left_index=True, right_index=True
    )

def combine_start_end(start_df, end_df):
    end_df = end_df.rename('end_timestamp')
    t0 = time.time()
    df = start_df.join(end_df, how='outer')
    df["n_days"] = df["end_timestamp"] - df["timestamp"]
    df["n_days"] = df["n_days"].apply(lambda x: x.days).fillna(0.0)

    # TODO : this bit is extremely slow!!
    df["range_dates"] = df.apply(
        lambda x: [
            x["timestamp"] + datetime.timedelta(days=day)
            for day in range(int(x.n_days))
        ],
        axis=1,
    )
    df = df[["age", "range_dates", "name_region"]].explode(column="range_dates")
    return df.rename(columns={"range_dates": "timestamp"})

def read_table_with_people(read, table, index, people_df, geography_df):
    df = read.table_to_df(table, index=index)
    df = df.merge(people_df, how="inner", left_index=True, right_index=True)
    if geography_df is not None:
        df = df.merge(geography_df, left_on="area_id", right_index=True, how="inner")
    if "timestamp" in df.columns:
        df["timestamp"] = pd.to_datetime(df["timestamp"])
    return df

def read_table_with_locations(
    read, table, index, people_df, geography_df, location_type="hospital"
):
    df = read_table_with_people(
        read, table, index=index, people_df=people_df, geography_df=geography_df
    )
    locations_df = read.table_to_df("locations", "id")
    locations_df = locations_df[locations_df["spec"] == location_type]
    locations_df.set_index("group_id", inplace=True)
    return df.merge(locations_df, left_on=f"{location_type}_ids", right_index=True)

def regional_outputs(logger_f: Union[Path, str], age_bins=(0, 12, 25, 65, 101),
        min_date= '2020-05-01',
        max_date= '2020-12-31',):
    read = RecordReader(logger_f)
    print("loading people...")
    people_df = read.table_to_df("population", index="id")
    print("loading geography...")
    geography_df = read.get_geography_df()
    geography_df = geography_df.drop_duplicates()
    print("loading infections...")
    infections_df = read_table_with_people(
        read,
        "infections",
        index="infected_ids",
        people_df=people_df,
        geography_df=geography_df,
    )
    print("loading deaths...")
    deaths_df = read_table_with_people(
        read,
        "deaths",
        index="dead_person_ids",
        people_df=people_df,
        geography_df=geography_df,
    )
    print("loading hospital admissions...")
    hosp_admissions_df = read_table_with_locations(
        read,
        "hospital_admissions",
        index="patient_ids",
        people_df=people_df,
        geography_df=geography_df,
    )
    print("loading icu admissions...")
    icu_admissions_df = read_table_with_locations(
        read,
        "icu_admissions",
        index="patient_ids",
        people_df=people_df,
        geography_df=geography_df,
    )
    print("loading discharges...")
    discharges_df = read_table_with_locations(
        read,
        "discharges",
        index="patient_ids",
        people_df=people_df,
        geography_df=geography_df,
    )
    print("loading recoveries...")
    recoveries_df = read_table_with_people(
        read,
        "recoveries",
        index="recovered_person_ids",
        people_df=people_df,
        geography_df=geography_df,
    )
    print("loading infection locations...")
    infection_locations = get_infection_locations(infections_df)
    print("loading regional infections...")
    regional_infections = get_regional_outputs(infections_df, age_bins, "infected")
    pd.testing.assert_series_equal(
        regional_infections["infected"],
        infection_locations.sum(axis=1),
        check_names=False,
    )
    print("loading regional deaths...")
    regional_deaths = get_regional_outputs(deaths_df, age_bins, "deaths")
    print("loading regional hospital admissions...")
    regional_admissions = get_regional_outputs(
        hosp_admissions_df, age_bins, "hospital_admissions"
    )
    print("loading regional icu admissions...")
    regional_icu_admissions = get_regional_outputs(
        icu_admissions_df, age_bins, "icu_admissions"
    )
    print("loading regional hospital deaths...")
    hospital_deaths_df = deaths_df[deaths_df["location_specs"] == "hospital"]
    print("loading all discharges info...")
    all_discharges_df = hospital_deaths_df["timestamp"].append(
        discharges_df["timestamp"]
    )
    print("loading hospital in/out info...")
    hospital_in_out_df = combine_start_end(hosp_admissions_df, all_discharges_df,)
    print("loading regional current in hospital...")
    regional_current_in_hospital = get_regional_outputs(
        hospital_in_out_df, age_bins, "currently_in_hospital"
    )
    print("loading all end (?) infection...")
    all_end_infection_df = deaths_df["timestamp"].append(recoveries_df["timestamp"])

    print("loading in out (?) infection. Takes a while...")
    start = time.time()
    infected_in_out_df = combine_start_end(infections_df, all_end_infection_df)
    print(f"\tTook {time.time() - start} seconds")

    print("loading regional infected...")
    regional_current_infected = get_regional_outputs(
        infected_in_out_df, age_bins, "currently_infected"
    )
    print("loading regional recovered...")
    regional_recovered = get_regional_outputs(recoveries_df, age_bins, "recovered")
    people_df = people_df.merge(
        geography_df, left_on="area_id", right_index=True, how="inner"
    )
    print("loading people by age...")
    people_by_age = split_by_age(
        people_df, age_bins, "people", group_on=["name_region"]
    )
    regional_current_susceptible = pd.DataFrame()
    idx = pd.date_range(start=min_date, end=max_date)
    multi_idx = pd.MultiIndex.from_product([list(regional_current_infected.index.levels[0]), 
        list(idx)],
        names=list(regional_current_infected.index.names))

    for column in people_by_age.columns:
        age_suffix = "_".join(column.split("_")[1:])
        infected_ever = regional_infections.groupby(level=0)[
            "infected_" + age_suffix
        ].cumsum()

        column_name = "currently_susceptible_" + age_suffix
        regional_current_susceptible[column_name] = people_by_age[column].sub(infected_ever)
        regional_current_susceptible = regional_current_susceptible.reindex(multi_idx)
        regional_current_susceptible = regional_current_susceptible.unstack(
             level=0
            ).bfill().ffill().stack()
        regional_current_susceptible = regional_current_susceptible.reorder_levels([1,0])

    regional_current_susceptible[
        "currently_susceptible"
    ] = regional_current_susceptible.sum(axis=1)

    print("Concatenating information...")
    output = pd.concat(
        [
            infection_locations,
            regional_infections,
            regional_deaths,
            regional_admissions,
            regional_icu_admissions,
            regional_current_in_hospital,
            regional_current_infected,
            regional_recovered,
            regional_current_susceptible,
        ],
        axis=1,
    ).fillna(0.0).astype(int)

    return output


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"init_available_projects": "00_Create Project.ipynb",
         "summarize_h5": "00_Create Project.ipynb",
         "pgrid_to_run_parameters": "00_Create Project.ipynb",
         "collect_statistics": "00_Create Project.ipynb",
         "fix_geojson": "00_Create Project.ipynb",
         "main": "00_Create Project.ipynb"}

modules = ["create_project.py"]

doc_url = "https://bhoov.github.io/JUNE-vis/"

git_url = "https://github.com/bhoov/JUNE-vis/tree/master/"

def custom_doc_links(name): return None
